/*
 * $Id: maxent.cpp,v 1.28 2006/08/21 17:30:38 tsuruoka Exp $
 */

#include "crf.h"
#include <cmath>
#include <cstdio>
#include <cfloat>

using namespace std;

const int MAX_LEN = 1000;
const string BOS_LABEL = "!BOS!";
const string EOS_LABEL = "!EOS!";
//const bool USE_BOS_EOS = true;  // need to fix load_from_file(). BOE and EOS should be put first.
const bool USE_BOS_EOS = false;  

const bool OUTPUT_MARGINAL_PROB = true;

//const double MIN_FEATURE_WEIGHT = 0.001; // prune features when saving

extern int push_stop_watch();

int
CRF_Model::BLMVMFunctionGradient(double *x, double *f, double *g, int n)
{
  const int nf = _fb.Size();

  if (_inequality_width > 0) {
    assert(nf == n/2);
    for (int i = 0; i < nf; i++) {
      _va[i] = x[i];
      _vb[i] = x[i + nf];
      _vl[i] = _va[i] - _vb[i];
    }
  } else {
    assert(nf == n);
    for (int i = 0; i < n; i++) {
      _vl[i] = x[i];
    }
  }
  
  double score = update_model_expectation();

  if (_inequality_width > 0) {
    for (int i = 0; i < nf; i++) {
      g[i]      = -(_vee[i] - _vme[i] - _inequality_width);
      g[i + nf] = -(_vme[i] - _vee[i] - _inequality_width);
    }
  } else {
    if (_sigma == 0) {
      for (int i = 0; i < n; i++) {
        g[i] = -(_vee[i] - _vme[i]);
      }
    } else {
      const double c = 1 / (_sigma * _sigma);
      for (int i = 0; i < n; i++) {
        g[i] = -(_vee[i] - _vme[i] - c * _vl[i]);
      }
    }
  }

  *f = -score;

  return 0;
}

int
CRF_Model::BLMVMLowerAndUpperBounds(double *xl,double *xu,int n)
{
  if (_inequality_width > 0) {
    for (int i = 0; i < n; i++){
      xl[i] = 0;
      xu[i] = 10000.0;
    }
    return 0;
  }

  for (int i = 0; i < n; i++){
    xl[i] = -10000.0;
    xu[i] = 10000.0;
    //    xl[i] = -2.0;
    //    xu[i] = 2.0;
  }
  return 0;
}

int
CRF_Model::perform_LMVM()
{
  cerr << "performing LMVM" << endl;

  //  push_stop_watch();
  if (_inequality_width > 0) {
    int nvars = _fb.Size() * 2;
    double *x = (double*)malloc(nvars*sizeof(double)); 

    // INITIAL POINT
    for (int i = 0; i < nvars / 2; i++) {
      x[i] = _va[i];
      x[i + _fb.Size()] = _vb[i];
    }

    int info = BLMVMSolve(x, nvars);

    for (int i = 0; i < nvars / 2; i++) {
      _va[i] = x[i];
      _vb[i] = x[i + _fb.Size()];
      _vl[i] = _va[i] - _vb[i];
    }
  
    free(x);

    return 0;
  }

  int nvars = _fb.Size();
  double *x = (double*)malloc(nvars*sizeof(double)); 

  // INITIAL POINT
  for (int i = 0; i < nvars; i++) { x[i] = _vl[i]; }

  int info = BLMVMSolve(x, nvars);

  for (int i = 0; i < nvars; i++) { _vl[i] = x[i]; }
  
  free(x);

  return 0;
}


static int num_tags;
static double state_weight[MAX_LEN][CRF_Model::MAX_LABEL_TYPES];
static double edge_weight[CRF_Model::MAX_LABEL_TYPES][CRF_Model::MAX_LABEL_TYPES];
static double forward_cache[MAX_LEN][CRF_Model::MAX_LABEL_TYPES];
static double backward_cache[MAX_LEN][CRF_Model::MAX_LABEL_TYPES];
static int edge_feature_id[CRF_Model::MAX_LABEL_TYPES][CRF_Model::MAX_LABEL_TYPES];
static int backward_pointer[MAX_LEN][CRF_Model::MAX_LABEL_TYPES];



static double forward_prob(const int len)
{
  for (int x = 0; x < len; x++) {
    //    double maxv = 0;
    double total = 0;
    for (int i = 0; i < num_tags; i++) {
      double sum;
      if (x == 0) {
	sum = edge_weight[num_tags][i]; // BOS
      } else {
	sum = 0;
	for (int j = 0; j < num_tags; j++) {
	  sum += edge_weight[j][i] * forward_cache[x-1][j];
	}
      }
      sum *= state_weight[x][i];
      forward_cache[x][i] = sum;
      //      maxv = max(sum, maxv);
      total += sum;
    }
    //    maxv *= 0.0000000000001;
    for (int i = 0; i < num_tags; i++) {
      forward_cache[x][i] /= total;
      state_weight[x][i] /= total;
    }
  }
  double total = 0;
  for (int i = 0; i < num_tags; i++) {
    total += forward_cache[len-1][i] * edge_weight[i][num_tags+1]; // EOS
  }

  return total;
}


static double backward_prob(const int len)
{
  for (int x = len - 1; x >= 0; x--) {
    for (int i = 0; i < num_tags; i++) {
      double sum;
      if (x == len - 1) {
	sum = edge_weight[i][num_tags+1]; // EOS
      } else {
	sum = 0;
	for (int j = 0; j < num_tags; j++) {
	  sum += edge_weight[i][j] * backward_cache[x+1][j];
	}
      }
      sum *= state_weight[x][i];
      backward_cache[x][i] = sum;
    }
  }
  double total = 0;
  for (int i = 0; i < num_tags; i++) {
    total += backward_cache[0][i] * edge_weight[num_tags][i]; // BOS
  }

  return total;
}


void CRF_Model::initialize_edge_weights() const
{
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const int id = edge_feature_id[i][j];
      //      if (id < 0) { edge_weight[i][j] = 1; continue; }
      assert(id >= 0);
      const double ew = _vl[id];
      edge_weight[i][j] = exp(ew);
    }
  }


}


void CRF_Model::initialize_state_weights(const Sequence & seq) const
{
  for (int i = 0; i < seq.vs.size(); i++) {
    vector<double> powv(_num_classes, 0.0);
    const Sample & s = seq.vs[i];
    for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
      for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	const double w = _vl[*k];
	powv[_fb.Feature(*k).label()] += w;
      }
    }
    
    for (int j = 0; j < _num_classes; j++) {
      state_weight[i][j] = exp(powv[j]);
    }
  }
}

double CRF_Model::forward_backward(const Sequence & seq) const
{
  initialize_state_weights(seq);

  num_tags = _num_classes;

  double fp = forward_prob(seq.vs.size());
  double bp = backward_prob(seq.vs.size());
  assert(abs(fp - 1) < 0.01);
  assert(abs(bp - 1) < 0.01);
  /*
  if (seq.vs.size() > 60) {
    cerr << "len = " << seq.vs.size() << " ";
    cerr << "fp = " << fp << " ";
    cerr << "bp = " << bp << endl;
    for (int i = 0; i < seq.vs.size(); i++) {
      double m = 0;
      for (int j = 0; j < _num_classes; j++) {
	m = max(m, backward_cache[i][j]);
      }
      cerr << m << " ";
    }
    cerr << endl;
  }
  */
  /*
  if (!(fp > 0 && fp < DBL_MAX)) {
    cout << endl;
    cout << "fp = " << fp << endl;
    cout << "scaling factor " << endl;
    for (int i = 0; i < seq.vs.size(); i++) {
      cout << i << " " << scaling_factor[i] << endl;
    }
    cout << endl;

    for (int i = 0; i < seq.vs.size(); i++) {
      for (int j = 0; j < num_tags; j++) {
	cout << forward_cache[i][j] << "\t";
      }
      cout << endl;
    }
    //    forward_prob_recur(seq.vs.size(), seq.vs.size() - 2, 0);

    save_to_file("model");
    exit(1);
  }
  */
  if (!(fp > 0 && fp < DBL_MAX && bp > 0 && bp < DBL_MAX)) {
    cerr << endl << "error: floating overflow.  a different value of Gaussian prior might work." << endl;
    //cerr << endl << "error: floating overflow.  " << endl;
    exit(1);
  }

  //  assert(abs(fp - bp) < 0.1);
  assert(fp > 0 && fp < DBL_MAX);
  assert(bp > 0 && bp < DBL_MAX);

  return fp;
}



double CRF_Model::viterbi(const Sequence & seq, vector<int> & best_seq) const
{
  initialize_state_weights(seq);

  num_tags = _num_classes;

  const int len = seq.vs.size();

  for (int x = 0; x < len; x++) {
    double total = 0;
    for (int i = 0; i < num_tags; i++) {
      //      double m = DBL_MIN;
      double m = -DBL_MAX;
      if (x == 0) {
	m = edge_weight[num_tags][i]; // BOS
      } else {
	for (int j = 0; j < num_tags; j++) {
	  double score = edge_weight[j][i] * forward_cache[x-1][j];
	  if (score > m) {
	    m = score;
	    backward_pointer[x][i] = j;
	  }
	}
      }
      m *= state_weight[x][i];
      forward_cache[x][i] = m;
      total += m;
    }
    for (int i = 0; i < num_tags; i++) {
      forward_cache[x][i] /= total;
    }
  }

  //  double m = DBL_MIN;
  double m = -DBL_MAX;
  for (int i = 0; i < num_tags; i++) {
    double score = forward_cache[len-1][i] * edge_weight[i][num_tags+1]; // EOS
    if (score > m) {
      m = score;
      best_seq[len-1] = i;
    }
  }
  for (int x = len - 2; x >= 0; x--) {
    best_seq[x] = backward_pointer[x+1][best_seq[x+1]];
  }

  return 0;
}

double CRF_Model::calc_likelihood(const Sequence & seq, double fp)
{
  assert(fp == 1.0);
  const int len = seq.vs.size();

  double p = 1.0;
  for (int i = 0; i < len; i++) {
    p *= state_weight[i][seq.vs[i].label];
  }
  for (int i = 0; i < len-1; i++) {
    p *= edge_weight[seq.vs[i].label][seq.vs[i+1].label];
  }
  p *= edge_weight[_num_classes][seq.vs[0].label];       // BOS
  p *= edge_weight[seq.vs[len-1].label][_num_classes+1]; // EOS
  
  //  p /= fp;

  return p;
}


int
CRF_Model::make_feature_bag(const int cutoff)
{
#ifdef USE_HASH_MAP
  typedef __gnu_cxx::hash_map<unsigned int, int> map_type;
#else    
  typedef std::map<unsigned int, int> map_type;
#endif

  for (std::vector<Sequence>::const_iterator k = _vs.begin(); k != _vs.end(); k++) {
    for (std::vector<Sample>::const_iterator i = k->vs.begin(); i != k->vs.end(); i++) {
      for (std::vector<int>::const_iterator j = i->positive_features.begin(); j != i->positive_features.end(); j++) {
	const ME_Feature feature(i->label, *j);
	int id = _fb.Put(feature);
      }
    }
  }

  int c = 0;
  
  init_feature2mef();
  
  return 0;
}

double
CRF_Model::heldout_likelihood()
{
  double logl = 0;
  int ncorrect = 0, total_len = 0;

  initialize_edge_weights();
  for (std::vector<Sequence>::const_iterator i = _heldout.begin(); i != _heldout.end(); i++) {
    total_len += i->vs.size();
    double fp = forward_backward(*i);
    double p = calc_likelihood(*i, fp);
    logl += log(p);

    for (int j = 0; j < i->vs.size(); j++) {
      const Sample & s = i->vs[j];
      vector<double> wsum = calc_state_weight(*i, j);
      if (s.label == max_element(wsum.begin(), wsum.end()) - wsum.begin()) ncorrect++;
    }

  }
  _heldout_error = 1 - (double)ncorrect / total_len;

  return logl /= _heldout.size();

}

vector<double> CRF_Model::calc_state_weight(const Sequence & seq, const int i) const
{
  vector<double> wsum(_num_classes);
  for (int j = 0; j < _num_classes; j++) {
    wsum[j] = forward_cache[i][j] / state_weight[i][j] * backward_cache[i][j];
  }

  return wsum;
}

double
CRF_Model::update_model_expectation()
{
  double logl = 0;
  int ncorrect = 0, total_len = 0;

  _vme.resize(_fb.Size());
  for (int i = 0; i < _fb.Size(); i++) _vme[i] = 0;

  initialize_edge_weights();
  for (vector<Sequence>::const_iterator n = _vs.begin(); n != _vs.end(); n++) {
    const Sequence & seq = *n;
    total_len += seq.vs.size();

    double fp = forward_backward(seq);
    assert(fp == 1.0);
    double p = calc_likelihood(seq, fp);
    logl += log(p);

    for (int i = 0; i < seq.vs.size(); i++) {
      const Sample & s = seq.vs[i];

      // model expectation (state)
      vector<double> wsum = calc_state_weight(seq, i);
      for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
	for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	  //_vme[*k] += wsum[_fb.Feature(*k).label()] / fp;
	  _vme[*k] += wsum[_fb.Feature(*k).label()];
	}
      }
      if (s.label == max_element(wsum.begin(), wsum.end()) - wsum.begin()) ncorrect++;

      // model expectation (edge)
      if (i == seq.vs.size() - 1) continue;
      for (int j = 0; j < _num_classes; j++) {
	for (int k = 0; k < _num_classes; k++) {
	  const double lhs = forward_cache[i][j];
	  const double rhs = backward_cache[i+1][k];
	  assert(lhs != DBL_MAX && rhs != DBL_MAX);
	  //const double w = lhs * edge_weight[j][k] * rhs / fp;
	  const double w = lhs * edge_weight[j][k] * rhs;
	  _vme[edge_feature_id[j][k]] += w;
	}
      }
    }
    if (USE_BOS_EOS) {
      // model expectation (BOS -> *)
      for (int j = 0; j < _num_classes; j++) {
	const double rhs = backward_cache[0][j];
	//const double w = edge_weight[_num_classes][j] * rhs / fp;
	const double w = edge_weight[_num_classes][j] * rhs;
	_vme[edge_feature_id[_num_classes][j]] += w;
      }
      // model expectation (* -> EOS)
      const int len = seq.vs.size();
      for (int j = 0; j < _num_classes; j++) {
	const double lhs = forward_cache[len-1][j];
	//const double w = edge_weight[j][_num_classes+1] * lhs / fp;
	const double w = edge_weight[j][_num_classes+1] * lhs;
	_vme[edge_feature_id[j][_num_classes+1]] += w;
      }
    }
  }



  for (int i = 0; i < _fb.Size(); i++) {
    _vme[i] /= _vs.size();
  }
  
  _train_error = 1 - (double)ncorrect / total_len;

  logl /= _vs.size();
  
  if (_inequality_width > 0) {
    for (int i = 0; i < _fb.Size(); i++) {
      logl -= (_va[i] + _vb[i]) * _inequality_width;
    }
  } else {
    if (_sigma > 0) {
      const double c = 1/(2*_sigma*_sigma);
      for (int i = 0; i < _fb.Size(); i++) {
        logl -= _vl[i] * _vl[i] * c;
      }
    }
  }

  //logl /= _vs.size();
  
  //  fprintf(stderr, "iter =%3d  logl = %10.7f  train_acc = %7.5f\n", iter, logl, (double)ncorrect/train.size());
  //  fprintf(stderr, "logl = %10.7f  train_acc = %7.5f\n", logl, (double)ncorrect/_train.size());

  return logl;
}

void
CRF_Model::add_training_sample(const CRF_Sequence & seq)
{
  if (seq.vs.size() >= MAX_LEN) {
    cerr << "error: too long sequence";
    exit(1);
  }

  Sequence s1;
  for (vector<CRF_State>::const_iterator i = seq.vs.begin(); i != seq.vs.end(); i++) {
    if (i->label == BOS_LABEL || i->label == EOS_LABEL) {
      cerr << "error: the label name \"" << i->label << "\" is reserved. Use a different name.";
      exit(1);
    }
    Sample s;
    s.label = _label_bag.Put(i->label);
    if (s.label > MAX_LABEL_TYPES-2) {
      cerr << "error: too many types of labels." << endl;
      exit(1);
    }
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      s.positive_features.push_back(_featurename_bag.Put(*j));
    }
    s1.vs.push_back(s);
  }
  _vs.push_back(s1);
}

int
CRF_Model::train(const int cutoff,
		 const double sigma, const double widthfactor)
		 //		 const double Nsigma2, const double widthfactor)
{
  if (sigma > 0 && widthfactor > 0) {
  //  if (Nsigma2 > 0 && widthfactor > 0) {
    cerr << "error: Gausian prior and inequality modeling cannot be used together." << endl;
    return 0;
  }
  if (_vs.size() == 0) {
    cerr << "error: no training data." << endl;
    return 0;
  }
  if (_nheldout >= _vs.size()) {
    cerr << "error: too much heldout data. no training data is available." << endl;
    return 0;
  }
  //if (_nheldout > 0) random_shuffle(_vs.begin(), _vs.end());

  _label_bag.Put(BOS_LABEL);
  _label_bag.Put(EOS_LABEL);
  _num_classes = _label_bag.Size() - 2;
  
  for (int i = 0; i < _nheldout; i++) {
    _heldout.push_back(_vs.back());
    _vs.pop_back();
  }
  
  int total_len = 0;
  for (int i = 0; i < _vs.size(); i++) total_len += _vs[i].vs.size();
  
  //  _sigma = sqrt((double)total_len / Nsigma2);
  //  if (Nsigma2 == 0) _sigma = 0;
  _sigma = sigma;
  _inequality_width = widthfactor / _vs.size();
  
  if (cutoff > 0) cerr << "cutoff threshold = " << cutoff << endl;
  if (_sigma > 0) cerr << "Gaussian prior sigma = " << _sigma << endl;
    //    cerr << "N*sigma^2 = " << Nsigma2 << " sigma = " << _sigma << endl;
  if (widthfactor > 0)
    cerr << "widthfactor = " << widthfactor << endl;
  cerr << "preparing for estimation...";
  int C = make_feature_bag(cutoff);
  //  _vs.clear();
  cerr << "done" << endl;
  cerr << "number of state types = " << _num_classes << endl;
  cerr << "number of samples = " << _vs.size() << endl;
  cerr << "number of features = " << _fb.Size() << endl;

  cerr << "calculating empirical expectation...";
  _vee.resize(_fb.Size());
  for (int i = 0; i < _fb.Size(); i++) {
    _vee[i] = 0;
  }

  for (vector<Sequence>::const_iterator n = _vs.begin(); n != _vs.end(); n++) {
    const Sequence & seq = *n;
    for (int i = 0; i < seq.vs.size(); i++) {
      for (vector<int>::const_iterator j = seq.vs[i].positive_features.begin(); j != seq.vs[i].positive_features.end(); j++){
	for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	  if (_fb.Feature(*k).label() == seq.vs[i].label) _vee[*k] += 1.0;
	}
      }
    }
    for (int i = 0; i < seq.vs.size()-1; i++) {
      const int c0 = seq.vs[i].label;
      const int c1 = seq.vs[i+1].label;
      _vee[edge_feature_id[c0][c1]] += 1.0;
    }
    if (USE_BOS_EOS) {
      _vee[edge_feature_id[_num_classes][seq.vs[0].label]] += 1.0;
      _vee[edge_feature_id[seq.vs[seq.vs.size()-1].label][_num_classes+1]] += 1.0;
    }
  }

  for (int i = 0; i < _fb.Size(); i++) {
    _vee[i] /= _vs.size();
  }
  cerr << "done" << endl;
  
  _vl.resize(_fb.Size());
  for (int i = 0; i < _fb.Size(); i++) _vl[i] = 0.0;
  if (_inequality_width > 0) {
    _va.resize(_fb.Size());
    _vb.resize(_fb.Size());
    for (int i = 0; i < _fb.Size(); i++) {
      _va[i] = 0.0;
      _vb[i] = 0.0;
    }
  }

  perform_LMVM();

  if (_inequality_width > 0) {
    int sum = 0;
    for (int i = 0; i < _fb.Size(); i++) {
      if (_vl[i] != 0) sum++;
    }
    cerr << "number of active features = " << sum << endl;
  }
  
}

void
CRF_Model::get_features(list< pair< pair<string, string>, double> > & fl)
{
  fl.clear();
  for (MiniStringBag::map_type::const_iterator i = _featurename_bag.begin();
       i != _featurename_bag.end(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      string label = _label_bag.Str(j);
      string history = i->first;
      int id = _fb.Id(ME_Feature(j, i->second));
      if (id < 0) continue;
      fl.push_back( make_pair(make_pair(label, history), _vl[id]) );
    }
  }

}

bool
CRF_Model::load_from_file(const string & filename)
{
  FILE * fp = fopen(filename.c_str(), "r");
  if (!fp) {
    cerr << "error: cannot open " << filename << "!" << endl;
    return false;
  }

  _vl.clear();
  _label_bag.Clear();
  _featurename_bag.Clear();
  _fb.Clear();
  char buf[1024];
  while(fgets(buf, 1024, fp)) {
    string line(buf);
    string::size_type t1 = line.find_first_of('\t');
    string::size_type t2 = line.find_last_of('\t');
    string classname = line.substr(0, t1);
    string featurename = line.substr(t1 + 1, t2 - (t1 + 1) );
    float lambda;
    string w = line.substr(t2+1);
    sscanf(w.c_str(), "%f", &lambda);
      
    const int label = _label_bag.Put(classname);
    const int feature = _featurename_bag.Put(featurename);
    _fb.Put(ME_Feature(label, feature));
    _vl.push_back(lambda);
  }

  // for zero-wight edges
  _label_bag.Put(BOS_LABEL);
  _label_bag.Put(EOS_LABEL);
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const string & label1 = _label_bag.Str(j);
      const int l1 = _featurename_bag.Put("->\t" + label1);
      const int id = _fb.Id(ME_Feature(i, l1));
      if (id < 0) {
	_fb.Put(ME_Feature(i, l1));
	_vl.push_back(0);
      }
    }
  }

  _num_classes = _label_bag.Size() - 2;

  init_feature2mef();
  initialize_edge_weights();

  fclose(fp);

  return true;
}

void
CRF_Model::init_feature2mef()
{
  _feature2mef.clear();
  for (int i = 0; i < _featurename_bag.Size(); i++) {
    vector<int> vi;
    for (int k = 0; k < _num_classes; k++) {
      int id = _fb.Id(ME_Feature(k, i));
      if (id >= 0) vi.push_back(id);
    }
    _feature2mef.push_back(vi);
  }

  //init edge feature_id[][]
  for (int i = 0; i < _label_bag.Size(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      const string & label1 = _label_bag.Str(j);
      const int l1 = _featurename_bag.Put("->\t" + label1);
      const int id = _fb.Put(ME_Feature(i, l1));

      edge_feature_id[i][j] = id;
    }
  }

}

bool
CRF_Model::save_to_file(const string & filename, const double th) const
{
  FILE * fp = fopen(filename.c_str(), "w");
  if (!fp) {
    cerr << "error: cannot open " << filename << "!" << endl;
    return false;
  }

  for (MiniStringBag::map_type::const_iterator i = _featurename_bag.begin();
       i != _featurename_bag.end(); i++) {
    for (int j = 0; j < _label_bag.Size(); j++) {
      string label = _label_bag.Str(j);
      string history = i->first;
      int id = _fb.Id(ME_Feature(j, i->second));
      if (id < 0) continue;
      if (_vl[id] == 0) continue; // ignore zero-weight features
      if (abs(_vl[id]) < th) continue; // cut off low-weight features
      fprintf(fp, "%s\t%s\t%f\n", label.c_str(), history.c_str(), _vl[id]);
    }
  }

  fclose(fp);

  return true;
}

void CRF_Model::decode_forward_backward(CRF_Sequence & s0, 
					vector< map<string, double> > & tagp) const
{
  Sequence seq;

  for (vector<CRF_State>::const_iterator i = s0.vs.begin(); i != s0.vs.end(); i++) {
    Sample s;
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      const int id = _featurename_bag.Id(*j);
      if (id >= 0) s.positive_features.push_back(id);
    }
    seq.vs.push_back(s);
  }
  
  tagp.clear();
  double fp = forward_backward(seq);
  for (int i = 0; i < seq.vs.size(); i++) {
    vector<double> wsum = calc_state_weight(seq, i);
    map<string, double> tp;
    if (OUTPUT_MARGINAL_PROB) {
      double sum = 0;
      for (vector<double>::const_iterator j = wsum.begin(); j != wsum.end(); j++) sum += *j;
      s0.vs[i].label = "";
      assert(abs(sum -1) < 0.01);
      double maxp = -1;
      string maxtag;
      for (int j = 0; j < wsum.size(); j++) {
	double p = wsum[j]/sum;
	if (p <= 0.001) continue;
	tp[_label_bag.Str(j).c_str()] = p;
	if (p > maxp) { maxp = p; maxtag = _label_bag.Str(j).c_str();}
      }
      tagp.push_back(tp);
      s0.vs[i].label = maxtag;
    } else {
      const int l = max_element(wsum.begin(), wsum.end()) - wsum.begin();
      s0.vs[i].label = _label_bag.Str(l);
    }
  }

}

void CRF_Model::decode_viterbi(CRF_Sequence & s0) const
{
  Sequence seq;

  for (vector<CRF_State>::const_iterator i = s0.vs.begin(); i != s0.vs.end(); i++) {
    Sample s;
    for (vector<string>::const_iterator j = i->features.begin(); j != i->features.end(); j++) {
      const int id = _featurename_bag.Id(*j);
      if (id >= 0) s.positive_features.push_back(id);
    }
    seq.vs.push_back(s);
  }
  
  vector<int> vs(seq.vs.size());
  double p = viterbi(seq, vs);
  for (int i = 0; i < seq.vs.size(); i++) {
    s0.vs[i].label = _label_bag.Str(vs[i]);
  }

}

/*
 * $Log: maxent.cpp,v $
 */

